@@ -2,6 +2,8 @@ import os
 import click
 import uuid
 
+from dotenv import dotenv_values
+
 from laai.hpo.mlflow_utils import terminate_parent_run, get_active_run_feature_set_id, get_active_run_params
 
 VERSION = "0.0.1"
@@ -9,6 +11,36 @@ IMAGE_NAME = "osipov/laai"
 IMAGE_TAG = "latest"
 IMAGE = f"{IMAGE_NAME}:{IMAGE_TAG}"
 
+def get_backend_provider(dojo = None, job = None):
+	if job:
+		job_config = {
+			**dotenv_values(f"{os.getcwd()}/.job/.{job}/env.sh")
+		}
+		dojo_config = {
+			**dotenv_values(f"{os.getcwd()}/.dojo/.{job_config['LAAI_DOJO']}/env.sh")
+		}
+		backend = dojo_config['LAAI_BACKEND']
+		provider = dojo_config['LAAI_PROVIDER']
+		return backend, provider		
+	elif dojo:
+		dojo_config = {
+			**dotenv_values(f"{os.getcwd()}/.dojo/.{dojo}/env.sh")
+		}
+		backend = dojo_config['LAAI_BACKEND']
+		provider = dojo_config['LAAI_PROVIDER']
+		return backend, provider
+	else:
+		return None
+
+def get_volumes(provider = None):
+	volumes = {f"{os.getcwd()}" : {"bind": "/workspace", "mode": "rw"},}
+	if provider == 'local':
+		volumes = {
+			**volumes, 
+			**{"/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"}}
+		}
+	return volumes
+
 def assert_aws_credentials():
 	assert 'AWS_ACCESS_KEY_ID' in os.environ, "Missing environment variable AWS_ACCESS_KEY_ID"
 	assert 'AWS_DEFAULT_REGION' in os.environ, "Missing environment variable AWS_DEFAULT_REGION"
@@ -31,7 +63,7 @@ def cli():
 	pass
 
 @cli.command("init")
-@click.option('--provider', required=True, type=click.Choice(['aws', 'azure', 'gcp'], case_sensitive=False))
+@click.option('--provider', required=True, type=click.Choice(['aws', 'azure', 'gcp', 'local'], case_sensitive=False))
 @click.option('--backend', default='docker', type=click.Choice(['docker', 'k8s'], case_sensitive=False), show_default=True)
 @click.option('--worker-instances', default='1', type=int, show_default=True)
 @click.option('--worker-instance-type', default='t3.micro', show_default=True)
@@ -50,75 +82,120 @@ def init(provider,
 			ssh_connection_timeout,
 			ssh_connection_attempts):			
 	"""Initialize a training dojo in a specified infrastructure provider."""
-	assert provider == 'aws', f"Provider {provider} is not publicly available in version {VERSION}"
-	assert backend == 'docker', f"Backend {backend} is not publicly available in version {VERSION}"
+	DojoCLI.init(provider, 
+			backend, 
+			worker_instances, 
+			worker_instance_type, 
+			manager_instances, 
+			manager_instance_type,
+			volume_size,
+			ssh_connection_timeout,
+			ssh_connection_attempts)
 
-	#check for the aws settings
-	if (provider == 'aws'):
-		assert_aws_credentials()
+class DojoCLI(click.MultiCommand):
+	@cli.command("init")
+	@click.option('--provider', required=True, type=click.Choice(['aws', 'azure', 'gcp', 'local'], case_sensitive=False))
+	@click.option('--backend', default='docker', type=click.Choice(['docker', 'k8s'], case_sensitive=False), show_default=True)
+	@click.option('--worker-instances', default='1', type=int, show_default=True)
+	@click.option('--worker-instance-type', default='t3.micro', show_default=True)
+	@click.option('--manager-instances', default='1', type=int, show_default=True)
+	@click.option('--manager-instance-type', default='t3.micro', show_default=True)
+	@click.option('--volume-size', default='8', type=int, show_default=True)
+	@click.option('--ssh-connection-timeout', default='5', type=int, show_default=True)
+	@click.option('--ssh-connection-attempts', default='1', type=int, show_default=True)
+	def init(provider, 
+				backend, 
+				worker_instances, 
+				worker_instance_type, 
+				manager_instances, 
+				manager_instance_type,
+				volume_size,
+				ssh_connection_timeout,
+				ssh_connection_attempts):			
+		"""Initialize a training dojo in a specified infrastructure provider."""
+		assert provider == 'aws	' or provider == 'local', f"Provider {provider} is not publicly available in version {VERSION}"
+		assert backend == 'docker', f"Backend {backend} is not publicly available in version {VERSION}"
 
-		# #drop the provider prefix for the instance type, e.g. aws:t3.micro -> t3.micro
-		# assert worker_instance_type[:len(provider) + 1] == provider + ":",\
-		# 	f"Worker instance type {worker_instance_type} is missing a prefix {provider}:"
-		# worker_instance_type = worker_instance_type[len(provider) + 1:]
+		#check for the aws settings
+		if (provider == 'aws'):
+			assert_aws_credentials()
 
-		# manager_instance_type = manager_instance_type[len(provider) + 1:]
+			# #drop the provider prefix for the instance type, e.g. aws:t3.micro -> t3.micro
+			# assert worker_instance_type[:len(provider) + 1] == provider + ":",\
+			# 	f"Worker instance type {worker_instance_type} is missing a prefix {provider}:"
+			# worker_instance_type = worker_instance_type[len(provider) + 1:]
 
-	print(f"Checking for the latest updates...", end='')
-	client = prepare_docker()
-	print('done')
+			# manager_instance_type = manager_instance_type[len(provider) + 1:]
 
-	dojo = os.environ['LAAI_DOJO'] if 'LAAI_DOJO' in os.environ else None
-	while not dojo or dojo[:1].isalpha() is False:
-		dojo = uuid.uuid4().hex[:8]
-	print(f"Using a dojo store at {os.getcwd()}/.dojo/.{dojo}")
+		print(f"Checking for the latest updates...", end='')
+		client = prepare_docker()
+		print('done')
+
+		dojo = os.environ['LAAI_DOJO'] if 'LAAI_DOJO' in os.environ else None
+		while not dojo or dojo[:1].isalpha() is False:
+			dojo = uuid.uuid4().hex[:8]
+		print(f"Using a dojo store at {os.getcwd()}/.dojo/.{dojo}")
 
-	print(f"Preparing a dojo using {backend} on {provider} with {worker_instances} worker(s) and {manager_instances} manager(s) ...")
-	container = client.containers.run(	image=IMAGE, 
-		command="/bin/bash -c '/opt/laai/dojo_init.sh'",
-		volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
 		environment = { "LAAI_DOJO": dojo, 
 						"LAAI_VERSION": VERSION,
 						"LAAI_BACKEND": backend,
 						"LAAI_PROVIDER": provider,
-						"LAAI_DOJO_CONNECT_TIMEOUT": ssh_connection_timeout,
-						"LAAI_DOJO_CONNECT_ATTEMPTS": ssh_connection_attempts,
-						"AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
-						"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
-						"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
-						"TF_VAR_instances": worker_instances,
-						"TF_VAR_worker_instance_type": worker_instance_type,
-						"TF_VAR_manager_instances": manager_instances,
-						"TF_VAR_manager_instance_type": manager_instance_type,
-						"TF_VAR_volume_size": volume_size,
-						},
-		auto_remove = True,
-		detach=True)
+						}
+
+		volumes = {f"{os.getcwd()}" : {"bind": "/workspace", "mode": "rw"},}
 		
-	logs = container.attach(stream = True, logs = True)
-	for s in logs:
-		print(s.decode('utf-8').strip())	
+		if provider == 'local':
+			print(f"Preparing a dojo using {backend} on {provider} ...")
+			volumes = {
+				**volumes, 
+				**{"/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"}}
+			}
+		else:
+			environment = { **environment, **{
+							"LAAI_DOJO_CONNECT_TIMEOUT": ssh_connection_timeout,
+							"LAAI_DOJO_CONNECT_ATTEMPTS": ssh_connection_attempts,
+							"AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
+							"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
+							"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
+							"TF_VAR_instances": worker_instances,
+							"TF_VAR_worker_instance_type": worker_instance_type,
+							"TF_VAR_manager_instances": manager_instances,
+							"TF_VAR_manager_instance_type": manager_instance_type,
+							"TF_VAR_volume_size": volume_size,}
+			}		
+			print(f"Preparing a dojo using {backend} on {provider} with {worker_instances} worker(s) and {manager_instances} manager(s) ...")
 
-	print(dojo)
+		container = client.containers.run(	image = IMAGE, 
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/dojo_init.sh'",
+			volumes = volumes,
+			environment = environment,
+			auto_remove = True,
+			detach=True)
+			
+		logs = container.attach(stream = True, logs = True)
+		for s in logs:
+			print(s.decode('utf-8').strip())	
+
+		print(dojo)
 
-class DojoCLI(click.MultiCommand):
 	@click.command()	
 	@click.argument('dojo')
 	def activate(dojo):
-		assert_aws_credentials()
 		print(f"Activating dojo {dojo}")
-		
+		backend, provider = get_backend_provider(dojo)
+		if 'aws' == provider:
+			assert_aws_credentials()
+
 		client = prepare_docker()
-		
+		volumes = get_volumes(provider)
+
 		container = client.containers.run(	
-			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/dojo_activate.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
+			image = IMAGE, 
+			command = f"/bin/bash -c '/opt/laai/{backend}/{provider}/dojo_activate.sh'",
+			volumes = volumes ,
 			environment = { "LAAI_DOJO": dojo, 
 							},
-			auto_remove = False,
+			auto_remove = True,
 			detach = True )
 
 		logs = container.attach(stream = True, logs = True)
@@ -126,15 +203,23 @@ class DojoCLI(click.MultiCommand):
 			print(s.decode('utf-8'))	
 
 	@click.command()	
-	def ls():		
+	@click.option('--reverse', '-r',  is_flag = True, default = False)
+	def ls(reverse):		
+		"""Output a list with the training dojo(s) in the current workspace.		
+		By default, the dojo that was created first in the workspace is listed first."""
+
 		client = prepare_docker()
 		
+		environment = {
+			"LAAI_DOJO_LS_REVERSE": "" if reverse else "-r"
+		}
+
 		container = client.containers.run(	
-			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/dojo_ls.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
-			auto_remove = False,
+			image = IMAGE, 
+			command=f"/bin/bash -c '/opt/laai/dojo_ls.sh'",
+			volumes = get_volumes() ,
+			environment = environment,
+			auto_remove = True,
 			detach = True )
 
 		logs = container.attach(stream = True, logs = True)
@@ -145,12 +230,11 @@ class DojoCLI(click.MultiCommand):
 	def prune():
 		client = prepare_docker()
 		container = client.containers.run(	
-			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/dojo_prune.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
+			image = IMAGE, 
+			command=f"/bin/bash -c '/opt/laai/dojo_prune.sh'",
+			volumes = get_volumes() ,
 			auto_remove = True,
-			detach=True )
+			detach = True )
 
 		logs = container.attach(stream = True, logs = True)
 		for s in logs:
@@ -161,22 +245,29 @@ class DojoCLI(click.MultiCommand):
 	@click.command()
 	@click.argument('dojo')
 	def rm(dojo):
-		assert_aws_credentials()
 		print(f"Removing dojo {dojo} and deactivating store {os.getcwd()}/.dojo/.{dojo}")
+
+		backend, provider = get_backend_provider(dojo)
+		environment = {}
+		if 'aws' == provider:
+			assert_aws_credentials()
+			environment = {
+				"AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
+				"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
+				"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],				
+			}
 		
 		client = prepare_docker()
+
+		volumes = get_volumes(provider)
 		
 		container = client.containers.run(	
-			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/dojo_rm.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
-			environment = { "AWS_ACCESS_KEY_ID": os.environ['AWS_ACCESS_KEY_ID'],
-							"AWS_DEFAULT_REGION": os.environ['AWS_DEFAULT_REGION'],
-							"AWS_SECRET_ACCESS_KEY": os.environ['AWS_SECRET_ACCESS_KEY'],
-							"LAAI_DOJO": dojo, 
-							},
-			auto_remove = False,
+			image = IMAGE, 
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/dojo_rm.sh'",
+			volumes = volumes,
+			environment = { **{"LAAI_DOJO": dojo,}, 
+							**environment,},
+			auto_remove = True,
 			detach=True )
 
 		logs = container.attach(stream = True, logs = True)
@@ -187,31 +278,36 @@ class DojoCLI(click.MultiCommand):
 	@click.argument('dojo')
 	def inspect(dojo):	
 		"""Return key value settings for the specified dojo."""
+		backend, provider = get_backend_provider(dojo)
+		if 'aws' == provider:
+			assert_aws_credentials()
 
 		client = prepare_docker()
 		
 		container = client.containers.run(	
-			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/dojo_inspect.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
+			image = IMAGE, 
+			command=f"/bin/bash -c '/opt/laai/dojo_inspect.sh'",
+			volumes = get_volumes(provider) ,
 			environment = { "LAAI_DOJO": dojo, 
 							},
-			auto_remove = False,
-			detach=True )
+			auto_remove = True,
+			detach = True )
 
 		logs = container.attach(stream = True, logs = True)
 		for s in logs:
-			print(s.decode('utf-8'))	
+			s = s.decode('utf-8').strip()
+			if len(s):
+				print(s)
 
 	def list_commands(self, ctx):
-		return ['ls', 'rm', 'prune', 'activate', 'inspect']
+		return ['ls', 'rm', 'prune', 'activate', 'inspect', 'init']
 
 	def get_command(self, ctx, name):
 		return {"ls": DojoCLI.ls,
 				"rm": DojoCLI.rm,
 				"prune": DojoCLI.prune,
 				"inspect": DojoCLI.inspect,
+				"init": DojoCLI.init,
 				'activate': DojoCLI.activate}[name]
 
 @cli.command(cls=DojoCLI)
@@ -236,19 +332,20 @@ class JobCLI(click.MultiCommand):
 			net = ip_network(subnet)
 			assert ip_address(manager_ip) in net, f"Manager IP {manager_ip} is not in the CIDR subnet {subnet} range"
 
-		print(f"Checking for the latest version...", end='')
-		client = prepare_docker()
-		print('done')
-
 		job = None
 		while not job or job[:1].isalpha() is False:
 			job = uuid.uuid4().hex[:8]			
 		print(f"Creating a job {job} using image {image} in dojo {dojo}")
 
+		backend, provider = get_backend_provider(dojo)
+
+		print(f"Checking for the latest version...", end='')
+		client = prepare_docker()
+		print('done')
+
 		container = client.containers.run(image = IMAGE, 
-			command="/bin/bash -c '/opt/laai/job_create.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-												"mode": "rw"}} ,
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/job_create.sh'",
+			volumes = get_volumes(provider) ,
 			environment = { "LAAI_DOJO": dojo,
 							"LAAI_JOB": job,
 							"LAAI_JOB_IMAGE": image,
@@ -273,7 +370,7 @@ class JobCLI(click.MultiCommand):
 		
 		container = client.containers.run(	
 			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/job_inspect.sh'",
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/job_inspect.sh'",
 			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
 											"mode": "rw"}} ,
 			environment = { "LAAI_JOB": job, },
@@ -297,6 +394,8 @@ class JobCLI(click.MultiCommand):
 		client = prepare_docker()
 		print('done')
 
+		backend, provider = get_backend_provider(job = job)		
+
 		env = dict({ "LAAI_JOB": job,
 					 "LAAI_JOB_REPLICAS": replicas,
 					 "LAAI_JOB_CFG_DETACH": 1 if detach else 0,
@@ -304,10 +403,11 @@ class JobCLI(click.MultiCommand):
 					 **{"LAAI_JOB_ARG_" + k : v for k, v in env},
 					 **{"LAAI_JOB_PORT_" + k : v for k, v in port})
 
+		vol = get_volumes(provider)
+
 		container = client.containers.run(	image = IMAGE, 
-			command="/bin/bash -c '/opt/laai/job_start.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-												"mode": "rw"}} ,
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/job_start.sh'",
+			volumes = vol ,
 			environment = env,
 			auto_remove = True,
 			detach = True)
@@ -318,21 +418,28 @@ class JobCLI(click.MultiCommand):
 				print(s.decode('utf-8').strip())	
 
 	@click.command()	
-	def ls():		
-		"""List existing jobs in the current workspace."""
+	@click.option('--reverse', '-r',  is_flag = True, default = False)
+	def ls(reverse):		
+		"""List existing jobs in the current workspace.
+		By default, the dojo that was created first in the workspace is listed first."""
+
 		client = prepare_docker()
 		
+		env = {
+			"LAAI_JOB_LS_REVERSE": "" if reverse else "-r"
+		}
+		
 		container = client.containers.run(	
 			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/job_ls.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-											"mode": "rw"}} ,
+			command=f"/bin/bash -c '/opt/laai/job_ls.sh'",
+			environment = env,
+			volumes = get_volumes(),
 			auto_remove = False,
 			detach = True )
 
 		logs = container.attach(stream = True, logs = True)
 		for s in logs:
-			print(s.decode('utf-8').strip())	
+			print(s.decode('utf-8').strip())
 
 	@click.command()
 	@click.argument('job')
@@ -343,7 +450,7 @@ class JobCLI(click.MultiCommand):
 
 		container = client.containers.run(	
 			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/hpo_disable.sh'",
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/hpo_disable.sh'",
 			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
 											"mode": "rw"}} ,
 			environment = env,
@@ -356,7 +463,7 @@ class JobCLI(click.MultiCommand):
 		
 		container = client.containers.run(	
 			image=IMAGE, 
-			command="/bin/bash -c '/opt/laai/job_rm.sh'",
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/job_rm.sh'",
 			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
 											"mode": "rw"}} ,
 			environment = env,
@@ -444,6 +551,8 @@ class HpoCLI(click.MultiCommand):
 		client = prepare_docker()
 		print('done')
 
+		backend, provider = get_backend_provider(job = job)
+
 		env = dict({ "LAAI_JOB": job,
 					 "LAAI_HPO_IMAGE": image,
 					 "LAAI_HPO_JOB_RUNS": num_runs,
@@ -454,11 +563,11 @@ class HpoCLI(click.MultiCommand):
 					 **{"LAAI_JOB_ARG_" + k : v for k, v in env},
 					 **{"LAAI_JOB_PORT_" + k : v for k, v in port})
 
+		vol = get_volumes(provider)
 
 		container = client.containers.run(	image = IMAGE, 
-			command="/bin/bash -c '/opt/laai/hpo_enable.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-												"mode": "rw"}} ,
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/hpo_enable.sh'",
+			volumes = vol,
 			environment = env,
 			auto_remove = True,
 			detach = True)
@@ -477,12 +586,14 @@ class HpoCLI(click.MultiCommand):
 		client = prepare_docker()
 		print('done')
 
+		backend, provider = get_backend_provider(job = job)
+
 		env = dict({ "LAAI_JOB": job,})
+		vol = get_volumes(provider)
 
 		container = client.containers.run(	image = IMAGE, 
-			command="/bin/bash -c '/opt/laai/hpo_disable.sh'",
-			volumes = {f"{os.getcwd()}" : {"bind": "/workspace",
-												"mode": "rw"}} ,
+			command=f"/bin/bash -c '/opt/laai/{backend}/{provider}/hpo_disable.sh'",
+			volumes = vol,
 			environment = env,
 			auto_remove = True,
 			detach = True)
